#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

#include <hpic.h>

int main(int argc, char *argv[])
{
  char ch = '-';
  size_t i, j, k;

  char infile[HPIC_STRNL];
  char outfile[HPIC_STRNL];
  double outval = 1.0;
  size_t mapnum = 0;

  size_t nside;
  int order;
  int coord;
  int type;
  size_t nmaps;
  size_t npix;
  char command[HPIC_STRNL];
  char creator[HPIC_STRNL];
  char extname[HPIC_STRNL];
  char comment[HPIC_STRNL];

  hpic_int *pix = NULL;
  hpic_fltarr *maps = NULL;
  hpic_int *hits = NULL;
  hpic_float *errs = NULL;
  hpic_keys *keys;
  hpic_float *tempmap;
  hpic_float *outline = NULL;
  
  hpic_vec_index *nearest;
  float fpix;
  
  int temp;
  float val;

  if (argc == 1) {
    printf("\n");
    printf("<******************************************************>\n");
    printf("Usage:  hpic_outline -in <input FITS file>\n");
    printf("        -out <output FITS file>\n");
    printf("        -mapnum <map number to process>\n");
    printf("        -outval <set all edge pixels to this value>\n");
    printf("<******************************************************>\n\n");
    exit(1);
  }

  /* parse command line */

  for (i = 0; i < argc; i++) {
    if (argv[i][0] == ch) {
      if (!strcmp(argv[i], "-in")) {
        strncpy(infile, argv[i + 1], HPIC_STRNL);
        i++;
      }
      if (!strcmp(argv[i], "-out")) {
        strncpy(outfile, argv[i + 1], HPIC_STRNL);
        i++;
      }
      if (!strcmp(argv[i], "-mapnum")) {
        mapnum = (size_t) atoi(argv[i + 1]);
        i++;
      }
      if (!strcmp(argv[i], "-outval")) {
        outval = atof(argv[i + 1]);
        i++;
      }
    }
  }

  /* process FITS file */

  strcpy(comment, "");

  if (hpic_fits_map_test(infile, &nside, &order, &coord, &type, &nmaps)) {

    if (mapnum >= nmaps) {
      fprintf(stderr,
              "hpic_outline:  file contains %lu maps (columns), cannot access column %lu\n",
              nmaps, mapnum);
      exit(1);
    }

    keys = hpic_keys_alloc();
    maps = hpic_fltarr_alloc(nmaps);
    for (i = 0; i < nmaps; i++) {
      tempmap = hpic_float_alloc(nside, order, coord, HPIC_AUTO);
      hpic_fltarr_set(maps, i, tempmap);
    }
    outline = hpic_float_alloc(nside, order, coord, HPIC_AUTO);
    nearest = hpic_vec_index_alloc(0);

    if (type = HPIC_FITS_FULL) {
      hpic_fits_full_read(infile, creator, extname, maps, keys);
      tempmap = hpic_fltarr_get(maps, mapnum);
      for (i = 0; i < hpic_nside2npix(nside); i++) {
        val = hpic_float_get(tempmap, i);
        if (!hpic_is_fnull(val)) {
          hpic_neighbors(nside, order, i, nearest);
          for (j = 0; j < hpic_vec_index_n_get(nearest); j++) {
            fpix = hpic_float_get(tempmap, hpic_vec_index_get(nearest, j));
            if (hpic_is_fnull(fpix)) {
              hpic_float_set(outline, i, (float)outval);
              break;
            }
          }
        }
      }

      hpic_float_free(tempmap);
      hpic_fltarr_set(maps, mapnum, outline);
      hpic_fits_full_write(outfile, "hpic_outline", "Threshold Map", comment,
                           maps, keys);
    } else {
      errs = hpic_float_alloc(nside, order, coord, HPIC_AUTO);
      hits = hpic_int_alloc(nside, order, coord, HPIC_AUTO);
      pix = hpic_int_alloc(nside, order, coord, HPIC_AUTO);

      hpic_fits_cut_read(infile, creator, extname, pix, hits, errs, maps,
                         keys);
      tempmap = hpic_fltarr_get(maps, mapnum);
      for (i = 0; i < hpic_nside2npix(nside); i++) {
        val = hpic_float_get(tempmap, i);
        if (!hpic_is_fnull(val)) {
          hpic_neighbors(nside, order, i, nearest);
          for (j = 0; j < hpic_vec_index_n_get(nearest); j++) {
            fpix = hpic_float_get(tempmap, hpic_vec_index_get(nearest, j));
            if (hpic_is_fnull(fpix)) {
              hpic_float_set(outline, i, (float)outval);
              break;
            }
          }
        }
      }

      hpic_float_free(tempmap);
      hpic_fltarr_set(maps, mapnum, outline);
      hpic_fits_cut_write(outfile, "hpic_outline", "Outline Map", comment,
                          pix, hits, errs, maps, keys);

      hpic_int_free(pix);
      hpic_int_free(hits);
      hpic_float_free(errs);
    }

    hpic_vec_index_free(nearest);
    hpic_keys_free(keys);
    for (i = 0; i < nmaps; i++) {
      tempmap = hpic_fltarr_get(maps, i);
      hpic_float_free(tempmap);
    }
    hpic_fltarr_free(maps);

  } else {
    fprintf(stderr, "hpic_outline:  FITS file is not a recognized type\n");
    exit(1);
  }

  return 0;
}
